#ifndef _CUTPLANE_H_
#define _CUTPLANE_H_

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef OLD_STL
#include <algo.h>
#include <stack.h>
#include <map.h>
#include <pair.h>
//PHG #include <hash_map.h>
#else
#include <algorithm>
#include <stack>
#include <map>
#include <utility>
//PHG #include <hash_map>
using std::min;
using std::max;
using std::map;
using std::pair;
#endif


#include <Face.h>
#include <Vector3D.h>

extern double orient3d(double *, double *, double *, double *);

//-------------------------------------------------------------------------------

struct HashFcnEdge : unary_function<pair<int,int>, size_t> {

  size_t operator()(const pair<int,int> &key) const {
    
    return key.first < key.second ? key.first : key.second;
    //return key.first+key.second;

  }

};

//-------------------------------------------------------------------------------

struct HashFcnFace : unary_function<Face, size_t> {

  size_t operator()(const Face &key) const {

    const int *nodeNum = key.getNodeNum();

    int res = nodeNum[0] < nodeNum[1] ? nodeNum[0] : nodeNum[1];

    return res < nodeNum[2] ? res : nodeNum[2];
    //return nodeNum[0]+nodeNum[1]+nodeNum[2];

    /*
    size_t res = 0;
    size_t len = sizeof(Face);
    const char *p = reinterpret_cast<const char *>(&key);

    while (len--) res = (res<<1)^*p++;
    return res;
    */
  }

};

//-------------------------------------------------------------------------------

typedef map<pair<int,int>, int> MapEdge;
typedef map<Face, int> MapFace;

/* PHG
typedef hash_map<pair<int,int>, int, HashFcnEdge> MapEdge;
typedef hash_map<Face, int, HashFcnFace> MapFace;
*/

//-------------------------------------------------------------------------------

class CutLine {

  Vec3D d;
  Vec3D e;
  Vec3D bb[2];

public:

  CutLine(Vec3D &_d, Vec3D &_e) : d(_d), e(_e) {
    bb[0] = d;
    bb[1] = d;
    for (int k=0; k<3; ++k) {
      bb[0][k] = min(bb[0][k], e[k]);
      bb[1][k] = max(bb[1][k], e[k]);
    }
  }
  ~CutLine() {}

  Vec3D* getBoundingBox() { return bb; }
  Vec3D getOrigin() { return d; }
  Vec3D getEnd() { return e; }

  int checkPiercePoint(Vec3D&, Vec3D&, Vec3D&);
  void constructPiercePoint(Vec3D&, Vec3D&, Vec3D&, double*, double*, double*);

};

//-------------------------------------------------------------------------------

inline
int CutLine::checkPiercePoint(Vec3D& a, Vec3D& b, Vec3D& c)
{

  double vol1 = orient3d(a, b, c, d);
  double vol2 = orient3d(a, b, c, e);
  double vol3;

  if (vol1 * vol2 > 0.0) return 10;          // [e,d] doesn't cross the plane
  if (vol1 == 0.0 && vol2 == 0.0) return 10; // [e,d] is in the plane
  if (vol1 == 0.0 && vol2 != 0.0) return 10; // d touches the plane
  if (vol1 != 0.0 && vol2 == 0.0) return 10; // e touches the plane

  vol1 = orient3d(d, a, b, e);
  vol2 = orient3d(d, b, c, e);
  vol3 = orient3d(d, c, a, e);

  if ((vol1 > 0.0 && vol2 > 0.0 && vol3 > 0.0) ||
      (vol1 < 0.0 && vol2 < 0.0 && vol3 < 0.0))
    return 0;
  else if (vol1 == 0.0) {
    if (vol2*vol3 > 0.0)
      return 1;
    else if (vol2 == 0)
      return -2;
    else if (vol3 == 0)
      return -1;
    else
      return 10;
  }
  else if (vol2 == 0.0) {
    if (vol1*vol3 > 0.0)
      return 2;
    else if (vol1 == 0)
      return -2;
    else if (vol3 == 0)
      return -3;
    else
      return 10;
  }
  else if (vol3 == 0.0) {
    if (vol1*vol2 > 0.0)
      return 3;
    else if (vol1 == 0)
      return -1;
    else if (vol2 == 0)
      return -3;
    else
      return 10;
  }

  return 10;

}

//-------------------------------------------------------------------------------

inline
void CutLine::constructPiercePoint(Vec3D& a, Vec3D& b, Vec3D& c,
				   double* s, double* r, double* t) 
{

    Vec3D ac = a - c;
    Vec3D cb = c - b;
    Vec3D cd = c - d;
    Vec3D ed = e - d;

    double coef0 = ac[2]*cb[1] - ac[1]*cb[2];
    double coef1 = ac[2]*cb[0] - ac[0]*cb[2];
    double coef2 = ac[1]*cb[0] - ac[0]*cb[1];
    
    double den = ed[0]*coef0 - ed[1]*coef1 + ed[2]*coef2;
    if (den == 0.0) {
      fprintf(stderr, "Error: vanishing denominator\n");
      exit(1);
    }
    double ooden = 1.0 / den;

    if (s)
      *s = ooden * (cd[0]*coef0 - cd[1]*coef1 + cd[2]*coef2);

    if (r)
      *r = ooden * (ed[0] * (cb[2]*cd[1] - cb[1]*cd[2]) -
		    ed[1] * (cb[2]*cd[0] - cb[0]*cd[2]) +
		    ed[2] * (cb[1]*cd[0] - cb[0]*cd[1]));

    if (t)
      *t = ooden * (ed[0] * (ac[2]*cd[1] - ac[1]*cd[2]) -
		    ed[1] * (ac[2]*cd[0] - ac[0]*cd[2]) +
		    ed[2] * (ac[1]*cd[0] - ac[0]*cd[1]));

}

//-------------------------------------------------------------------------------

class CutPlane {

  Vec3D a;
  Vec3D b;
  Vec3D c;

  Vec3D normal;
  Vec3D bb[2];

public:

  CutPlane(Vec3D &, Vec3D &, Vec3D &);
  ~CutPlane() {}

  void computeBoundingBox(Vec3D [2]);
  int checkPiercePoint(Vec3D &, Vec3D &);
  double constructPiercePoint(Vec3D &, Vec3D &);

  Vec3D *getBoundingBox() { return bb; }
  Vec3D &getNormal() { return normal; }

};

//-------------------------------------------------------------------------------

inline
int CutPlane::checkPiercePoint(Vec3D &d, Vec3D &e)
{

  double vol1 = orient3d(a, b, c, d);
  double vol2 = orient3d(a, b, c, e);

  if (vol1 * vol2 > 0.0) return 0;               // [e,d] doesn't cross the plane
  else if (vol1 == 0.0 && vol2 != 0.0) return 2; // d touches the plane
  else if (vol1 != 0.0 && vol2 == 0.0) return 3; // e touches the plane
  else if (vol1 == 0.0 && vol2 == 0.0) return 4; // [e,d] is in the plane
  else return 1;                                 // [e,d] crosses the plane

}

//-------------------------------------------------------------------------------

inline
double CutPlane::constructPiercePoint(Vec3D &d, Vec3D &e)
{

  Vec3D p, ac, cb, cd, ed;

  ac = a - c;
  cb = c - b;
  cd = c - d;
  ed = e - d;

  double coef0 = ac[2]*cb[1] - ac[1]*cb[2];
  double coef1 = ac[2]*cb[0] - ac[0]*cb[2];
  double coef2 = ac[1]*cb[0] - ac[0]*cb[1];

  double den = ed[0] * coef0 - ed[1] * coef1 + ed[2] * coef2;

  if (den == 0.0) {
    fprintf(stderr, "vanishing denominator\n");
    exit(1);
  }

  double s = cd[0] * coef0 - cd[1] * coef1 + cd[2] * coef2;

  /*
  double r = (e[0]-d[0]) * ((c[2]-b[2])*(c[1]-d[1]) - (c[1]-b[1])*(c[2]-d[2])) -
             (e[1]-d[1]) * ((c[2]-b[2])*(c[0]-d[0]) - (c[0]-b[0])*(c[2]-d[2])) +
             (e[2]-d[2]) * ((c[1]-b[1])*(c[0]-d[0]) - (c[0]-b[0])*(c[1]-d[1]));
  r /= den;

  double t = (e[0]-d[0]) * ((a[2]-c[2])*(c[1]-d[1]) - (a[1]-c[1])*(c[2]-d[2])) -
             (e[1]-d[1]) * ((a[2]-c[2])*(c[1]-d[1]) - (a[0]-c[0])*(c[2]-d[2])) +
             (e[2]-d[2]) * ((a[1]-c[1])*(c[0]-d[0]) - (a[0]-c[0])*(c[1]-d[1]));
  t /= den;
  */

  s /= den;

  //p = d + s * (e - d);

  //Vec3D pp = c + r * (a - c) - t * (c - b);

  return s;

}

//-------------------------------------------------------------------------------

#endif
