#include <Domain.h>
#include <CutPlane.h>
#include <Vector.h>
#include <GetTime.h>
#include <BinFileHandler.h>

#ifdef USE_ANN
#include <ANN/ANN.h>
#endif

#include <stdio.h>
#include <math.h>
#include <alloca.h>
#include <dlfcn.h>

#include <list>
#include <algorithm>
#include <iostream>

extern bool gmsh;
extern bool symmetry;
extern double xyzsym[3];
extern void checkFileOpen(FILE *, char *);

//-------------------------------------------------------------------------------

void Domain::computeBoundingBox(SVec<double,3> &X, bool info)
{

  int k;
  for (k=0; k<3; ++k) {
    bb[0][k] = X[0][k];
    bb[1][k] = X[0][k];
  }

  int loc[2][3] = {{-1,-1,-1}, {-1,-1,-1}};

  int i;
  for (i=0; i<X.size(); ++i) {
    for (k=0; k<3; ++k) {
      if (X[i][k] < bb[0][k]) { 
	bb[0][k] = X[i][k]; 
	loc[0][k] = i; 
      }
      if (X[i][k] > bb[1][k]) { 
	bb[1][k] = X[i][k]; 
	loc[1][k] = i; 
      }
    }
  }

  if (info) {
    fprintf(stdout, "mesh bb min loc: %d %d %d\n", 
	    loc[0][0]+1, loc[0][1]+1, loc[0][2]+1);
    fprintf(stdout, "mesh bb max loc: %d %d %d\n", 
	    loc[1][0]+1, loc[1][1]+1, loc[1][2]+1);
    for (k=0; k<3; ++k) {
      fprintf(stdout, "node %d = %e %e %e\n", 
	      loc[0][k]+1, X[ loc[0][k] ][0], X[ loc[0][k] ][1], X[ loc[0][k] ][2]);
      fprintf(stdout, "node %d = %e %e %e\n", 
	      loc[1][k]+1, X[ loc[1][k] ][0], X[ loc[1][k] ][1], X[ loc[1][k] ][2]);
    }
  }

}

//-------------------------------------------------------------------------------

void Domain::numberEdges() 	
{

  double t0 = getCpuTime();

  fprintf(stdout, " ... numbering"); fflush(stdout);

  tets.numberEdges(edges);

  fprintf(stdout, " %d edges for <%s> in %f s\n", edges.size(), name, getCpuTime()-t0);
  fflush(stdout);

}

//-------------------------------------------------------------------------------

//#define RENAULT
//#define DAVE

void Domain::linecuts(char* nsol, char* npos, char* suffix, CutLine* line)
{

  Vec<CutLinePt> points(0);
  SVec<double,3> X(nodes.size());

  int i, k;
  for (i=0; i<nodes.size(); ++i)
    for (k=0; k<3; ++k)
      X[i][k] = nodes[i][k];

  makelinecut(suffix, *line, X, points);

  if (nsol[0] != 0) {
    int dim = 0;
    if (postop)
      dim = postop->getNumberOfNodalSolutions(nsol);
    if (dim == 0) {
      fprintf(stderr, "*** Error: no valid solution\n");
      exit(1);
    }
    Vec<double> sol(dim*nodes.size());
    Vec<Vec3D> x(points.size());
    Vec<double> ul(points.size()*dim);
    double* u = new double[3*dim];
    FILE* fp = 0;
    int numsol = 1;
    if (postop)
      numsol = postop->getNumberOfSolutions(nsol);
    int isol;
#ifdef DAVE
    Vec<double> soln(dim*nodes.size());
    Vec<double> solnp1(dim*nodes.size());
    double tn;
    double tnp1;
    double t = 0.0;
    sol = 0.0;
    for (isol=0; isol<numsol; ++isol) {
      readSolution(nsol, fp, PostOp::ASCII, sol.size(), isol, tnp1, solnp1.data());
      if (isol > 0) {
	double dt = tnp1 - tn;
	//fprintf(stderr, "dt=%e\n", dt);
	t += dt;
	sol += 0.5 * dt * (soln + solnp1);
      }
      tn = tnp1;
      soln = solnp1;
    }
    sol *= 1.0 / t;
    numsol = 1;
#endif
    for (isol=0; isol<numsol; ++isol) {
      double tag = 0.0;
      bool status;
#ifdef DAVE
      status = true;
#else
      status = readSolution(nsol, fp, PostOp::ASCII, sol.size(), isol, tag, sol.data());
#endif
      if (status) {
	char name[MAXLINE];
	if (numsol > 1)
	  sprintf(name, "slice.gnu.%d%s", isol+1, suffix);
	else
	  sprintf(name, "slice.gnu%s", suffix);
	FILE* fgnu = fopen(name, "w");
	fprintf(fgnu, "# %g\n", tag);
#ifdef RENAULT
	dim = 1;
#endif
	for (i=0; i<points.size(); ++i) {
	  Vec3D p0(X[points[i][0]]);
	  Vec3D p1(X[points[i][1]]);
	  Vec3D p2(X[points[i][2]]);
	  x[i] = p2 + points[i].r * (p0 - p2) - points[i].t * (p2 - p1);
	  for (k=0; k<3; ++k) {
	    for (int j=0; j<dim; ++j) {
#ifdef RENAULT
	      u[k*dim+j] = sqrt(sol[points[i][k]*dim+1]*sol[points[i][k]*dim+1] +
				sol[points[i][k]*dim+2]*sol[points[i][k]*dim+2] + 
				sol[points[i][k]*dim+3]*sol[points[i][k]*dim+3]) / 
		sol[points[i][k]*dim+0];
#else
	      u[k*dim+j] = sol[points[i][k]*dim+j];
#endif
	    }
	  }
	  fprintf(fgnu, "%e %e %e %e", x[i][0], x[i][1], x[i][2], sqrt((x[i]-x[0])*(x[i]-x[0])));
	  for (int j=0; j<dim; ++j) {
	    ul[i*dim+j] = u[2*dim+j] + points[i].r * (u[0*dim+j] - u[2*dim+j]) - 
	      points[i].t * (u[2*dim+j] - u[1*dim+j]);
	    fprintf(fgnu, " %e", ul[i*dim+j]);
	  }
	  fprintf(fgnu, "\n");
      }
#ifdef RENAULT
	double um = 0.0;
	for (i=1; i<points.size(); ++i)
	  um += 0.5 * (ul[i]+ul[i-1]) * sqrt((x[i]-x[i-1])*(x[i]-x[i-1]));
	um /= sqrt((x[points.size()-1]-x[0])*(x[points.size()-1]-x[0]));
	fprintf(fgnu, "# Mean value = %e\n", um);
#endif
	fclose(fgnu);
      }
    if (fp) fclose(fp);
    }
  }
}

//-------------------------------------------------------------------------------

void Domain::makelinecut(char* suffix, CutLine& line, SVec<double,3> &X, Vec<CutLinePt>& points)
{

  map<int,int> lcnodes; 
  MapEdge lcedges;
  MapFace lcfaces;
  stack<CutLinePt> pts;
  MapEdge seg;

  int noverlaps = 0;
  int ncuts = 0;
  Vec3D lp0 = line.getOrigin();
  Vec3D* lbb = line.getBoundingBox();

  double t0 = getCpuTime();

  int i;
  for (i=0; i<tets.size(); ++i) {
    tets[i].computeBoundingBox(X);
    if (tets[i].checkOverlap(lbb)) {
      ++noverlaps;
      ncuts += tets[i].makelinecut(X, line, lcnodes, lcedges, lcfaces, pts, seg);
    }
  }

  t0 = getCpuTime() - t0;

  fprintf(stdout, "     CPU time = %f s\n", t0);
  fprintf(stdout, "     %g%% of the tetrahedra overlapped the line bounding boxes\n",
	  100.0*double(noverlaps)/double(tets.size()));
  fprintf(stdout, "     %g%s of the tetrahedra cut the line bounding boxes\n",
	  100.0*double(ncuts)/double(tets.size()), "%");

  int npoints = pts.size();
  int nsegments = seg.size();

  Vec<CutLinePt> ptsvec(npoints);
  i = npoints;
  while (pts.size()) {
    --i;
    ptsvec[i] = pts.top();
    pts.pop();
  }

  map<double,int> ptsmap;
  for (i=0; i<npoints; ++i) {
    Vec3D p0(X[ptsvec[i][0]]);
    Vec3D p1(X[ptsvec[i][1]]);
    Vec3D p2(X[ptsvec[i][2]]);
    Vec3D x = p2 + ptsvec[i].r * (p0 - p2) - ptsvec[i].t * (p2 - p1);
    double dist = sqrt((lp0 - x)*(lp0 - x));
    ptsmap[dist] = i;
  }

  points.resize(npoints);
  Vec<int> order(npoints);
  i = 0;
  for (map<double,int>::iterator pit=ptsmap.begin(); pit!=ptsmap.end(); ++pit) {
    points[i] = ptsvec[pit->second];
    order[pit->second] = i;
    ++i;
  }

  /*
  */
  SVec<int,2> segments(nsegments);
  for (MapEdge::iterator it=seg.begin(); it!=seg.end(); ++it) {
    segments[it->second][0] = order[it->first.first];
    segments[it->second][1] = order[it->first.second];
  }


  char name[MAXLINE];
  sprintf(name, "slice.top%s", suffix);
  FILE* ftop = fopen(name, "w");
  fprintf(ftop, "Nodes SliceNodes%s\n", suffix);
  for (i=0; i<npoints; ++i) {
    Vec3D p0(X[points[i][0]]);
    Vec3D p1(X[points[i][1]]);
    Vec3D p2(X[points[i][2]]);
    Vec3D x = p2 + points[i].r * (p0 - p2) - points[i].t * (p2 - p1);
    fprintf(ftop, "%d %g %g %g\n", i+1, x[0], x[1], x[2]);
  }
  fprintf(ftop, "typedef 3 Line_2\n");
  fprintf(ftop, "Elements SliceElements%s using SliceNodes%s\n", suffix, suffix);
  for (i=0; i<nsegments; ++i)
    fprintf(ftop, "%d 3 %d %d\n", i+1, segments[i][0]+1, segments[i][1]+1);
  if (ftop) fclose(ftop);

}

//-------------------------------------------------------------------------------

void Domain::planecuts(char *nameSol, char *namePos, char* suffix, 
                       CutPlane *cutPlane, int freq, int minplane, int maxplane)
{

  double cput0 = getCpuTime();
  double t0 = getTime();
  FILE *fsol1 = 0;
  FILE *fsol2 = 0;
  FILE *fxpost = 0;;
 
  //TG
  int solSize;
  if(postop && nameSol[0] != 0)
  {
	  solSize = nodes.size()*postop->getNumberOfNodalSolutions(nameSol);
	  solutionDimension = postop->getNumberOfNodalSolutions(nameSol);
  }
  else
	  solSize = nodes.size();
  
  SVec<double,3> X(nodes.size());
  Vec<double> sol(solSize);
  X = nodes;
  sol = 0.0;

  SVec<double,3> Xcut(0);
  SVec<int,4> Fcut(0);
  SVec<int,3> IJcut(0);
  Vec<double> Scut(0);
  Vec<int> Ncut(0);
  Vec<int> Tcut(0);

  if (gmsh) {
    char name[MAXLINE];
    sprintf(name, "slice2D.pos%s", suffix);
    fsol1 = fopen(name, "w");
    sprintf(name, "slice3D.pos%s", suffix);
    fsol2 = fopen(name, "w");
  }

  int i = 0;
  while (1) {
    double tag = 0.0;
    bool status = false;
    bool process = true;
    if(maxplane>0) process = i % freq == 0 && i >= minplane && i <= maxplane;
    else           process = i % freq == 0 && i >= minplane;

    if(process){
      if (namePos[0] != 0)
        status = readPosition(namePos, i, tag, X);
      if (nameSol[0] != 0)
      {
        status = readSolution(nameSol, fxpost, PostOp::ASCII, sol.size(), i, tag, sol.data());
      }
    }else status = true;
 
    if (i != 0 && !status) break;
    if (i > maxplane && maxplane > 0) break;

    if (i == 0 || namePos[0] != 0)
      makeCut(suffix, *cutPlane, X, Xcut, Fcut, IJcut, Scut, Ncut, Tcut);

    if (i == 0 && nameSol[0] != 0 && !gmsh) {
      char name[MAXLINE];
      sprintf(name, "slice2D.sol%s", suffix);
      fsol1 = fopen(name, "w");
      if(solutionDimension == 1) // TG
	      fprintf(fsol1, "Scalar scalar under load for SliceNodes2D%s\n", suffix);
      else
	      if(solutionDimension == 3)
	       		fprintf(fsol1, "Vector vector under load for SliceNodes2D%s\n", suffix);
      	      else
	      {
			fprintf(fsol1, "??? name under load for SliceNodes2D%s\n", suffix);
			fprintf(stderr,"Warning : This doesn't look like scalar or vector data, please check solution file header");
	      }
	
      fprintf(fsol1, "%d\n", Xcut.size());
      sprintf(name, "slice3D.sol%s", suffix);
      fsol2 = fopen(name, "w");
      if(solutionDimension == 1) // TG
	      fprintf(fsol2, "Scalar scalar under load for SliceNodes3D%s\n", suffix);
        else
	      if(solutionDimension == 3)
	       		fprintf(fsol2, "Vector vector under load for SliceNodes3D%s\n", suffix);

      	      else
	      {
		      fprintf(fsol2, "??? name under load for SliceNodes3D%s\n", suffix);
		      fprintf(stderr,"Warning : This doesn't look like scalar or vector data, please check solution file header");
	      }

      fprintf(fsol2, "%d\n", Ncut.size());
    }

    if(process){
      fprintf(stderr, "     Writing slice %d with tag %g\n", i+1, tag);
      constructTraceOnSurface("contour", suffix,Xcut, sol, Fcut, IJcut, Scut);
      writeCut(fsol1, fsol2, tag, X, sol, Xcut, Fcut, IJcut, Scut, Ncut, Tcut);
    }

    ++i;
  }

  if (fsol1) fclose(fsol1);
  if (fsol2) fclose(fsol2);
  if (fxpost) fclose(fxpost);

  fprintf(stdout, "Timings:\n");
  fprintf(stdout, "\t\tCPU Timing:\t%e s\n", getCpuTime()-cput0);
  fprintf(stdout, "\t\t    Timing:\t%e s\n", getTime()-t0);

}

//-------------------------------------------------------------------------------

void Domain::makeCut(char* suffix, CutPlane& cutPlane, SVec<double,3>& X, 
		     SVec<double,3>& Xcut, SVec<int,4>& Fcut, SVec<int,3>& IJcut, 
		     Vec<double>& Scut, Vec<int>& Ncut, Vec<int>& Tcut)
{

  double t0 = getCpuTime();

  Vec<int> nodeCut(nodes.size()); 
  MapEdge listCutEdges;
  MapFace listCutFaces;
  stack<CutPt> listCutPts;
  stack<int> listCutTets;

  nodeCut = -1;

  computeBoundingBox(X);  
  cutPlane.computeBoundingBox(bb);
  Vec3D *cpbb = cutPlane.getBoundingBox();

  int numOverlaps = 0;
  int numCuts = 0;

  int i, j;
  for (i=0; i<tets.size(); ++i) {
    tets[i].computeBoundingBox(X);
    if (tets[i].checkOverlap(cpbb)) {
      ++numOverlaps;
      int res = tets[i].makeplanecut(X, cutPlane, nodeCut, listCutEdges, listCutFaces, listCutPts);
      if (res == 1) {
	++numCuts;
	/*
	int jj = 0;
	for (j=0; j<4; ++j) {
	  if (sqrt( (nodes[tets[i][j]][0]-x0[0])*(nodes[tets[i][j]][0]-x0[0]) +
		    (nodes[tets[i][j]][1]-x0[1])*(nodes[tets[i][j]][1]-x0[1]) +
		    (nodes[tets[i][j]][2]-x0[2])*(nodes[tets[i][j]][2]-x0[2]) ) < radius)
	    ++jj;
	}
	if (jj == 4)
	*/
	listCutTets.push(i);
      }
    }
  }

  Tcut.resize(listCutTets.size());
  i = 0;
  while (listCutTets.size()) {
    Tcut[i++] = listCutTets.top();
    listCutTets.pop();
  }

  Vec<int> renum(nodes.size());
  renum = -1;
  int count = 0;
  for (i=0; i<Tcut.size(); ++i)
    for (j=0; j<4; ++j)
      if (renum[tets[Tcut[i]][j]] == -1)
	renum[tets[Tcut[i]][j]] = count++;
  Ncut.resize(count);
  for (i=0; i<nodes.size(); ++i)
    if (renum[i] != -1)
      Ncut[renum[i]] = i;

  if (!gmsh) {
    char name[MAXLINE];
    sprintf(name, "slice3D.top%s", suffix);
    FILE* ftop2 = fopen(name, "w");
    fprintf(ftop2, "Nodes SliceNodes3D%s\n", suffix);
    for (i=0; i<count; ++i)
      fprintf(ftop2, "%d %e %e %e\n", i+1, X[Ncut[i]][0], X[Ncut[i]][1], X[Ncut[i]][2]);
    fprintf(ftop2, "Elements SliceElements3D%s using SliceNodes3D%s\n", suffix, suffix);
    for (i=0; i<Tcut.size(); ++i)
      fprintf(ftop2, "%d 5 %d %d %d %d\n", i+1, renum[tets[Tcut[i]][0]]+1, 
	      renum[tets[Tcut[i]][1]]+1, renum[tets[Tcut[i]][2]]+1, renum[tets[Tcut[i]][3]]+1);
    fclose(ftop2);
  }

  int numPts = listCutPts.size();
  Xcut.resize(numPts);
  IJcut.resize(numPts);
  Scut.resize(numPts);

  nodeCut = 0;

  for (i=0; i<faces.size(); ++i)
    if (faces[i].getCode() < 0)
      for (int k=0; k<3; ++k)
	nodeCut[ faces[i][k] ] = 1;

  i = numPts;
  while (listCutPts.size()) {
    --i;
    CutPt pt = listCutPts.top();
    for (int k=0; k<3; ++k)
      Xcut[i][k] = X[pt.i][k] + pt.s * (X[pt.j][k] - X[pt.i][k]);
    IJcut[i][0] = pt.i;
    IJcut[i][1] = pt.j;
    if (nodeCut[pt.i] == 1 && nodeCut[pt.j] == 1)
      IJcut[i][2] = 1;
    else 
      IJcut[i][2] = 0;
    Scut[i] = pt.s;
    listCutPts.pop();
  }

  int typeFace[6]; 
  for (i=1; i<=5; ++i) 
    typeFace[i] = 0;

  int numFaces = listCutFaces.size();
  Fcut.resize(numFaces);

  i = 0;
  for (MapFace::iterator it = listCutFaces.begin(); it != listCutFaces.end(); ++it) {
    Face f = it->first;
    f.reorient(Xcut, cutPlane.getNormal());
    Fcut[i][0] = f[0];
    Fcut[i][1] = f[1];
    Fcut[i][2] = f[2];
    Fcut[i][3] = it->second;
    ++i;
    ++typeFace[ it->second ]; 
  }

  t0 = getCpuTime() - t0;

  fprintf(stdout, "     CPU time = %f s\n", t0);
  fprintf(stdout, "     %3.2f%s of the tetrahedra overlapped the plane bounding boxes\n",
	  100.0*double(numOverlaps)/double(tets.size()), "%");
  fprintf(stdout, "     %3.2f%s of the tetrahedra cut the plane bounding boxes\n",
	  100.0*double(numCuts)/double(tets.size()), "%");
  for (i=1; i<=5; ++i)
    fprintf(stdout, "     %d element(s) of type %d\n", typeFace[i], i);

  if (!gmsh) {
    char name[MAXLINE];
    sprintf(name, "slice2D.top%s", suffix);
    FILE *ftop = fopen(name, "w");
    fprintf(ftop, "Nodes SliceNodes2D%s\n", suffix);
    for (i=0; i<numPts; ++i)
      fprintf(ftop, "%d %g %g %g\n", i+1, Xcut[i][0], Xcut[i][1], Xcut[i][2]);
    fprintf(ftop, "Elements SliceElements2D%s using SliceNodes2D%s\n", suffix, suffix);
    for (i=0; i<numFaces; ++i)
      fprintf(ftop, "%d 4 %d %d %d\n", i+1, Fcut[i][0]+1, Fcut[i][1]+1, Fcut[i][2]+1);
    fclose(ftop);
  }

}

//-------------------------------------------------------------------------------

void Domain::writeCut(FILE* fp1, FILE* fp2, double tag, SVec<double,3>& X, Vec<double> &sol, 
		      SVec<double,3> &Xcut, SVec<int,4> &Fcut, SVec<int,3> &IJcut, 
		      Vec<double> &Scut, Vec<int>& Ncut, Vec<int>& Tcut)
{
  if (!gmsh && fp1) {
    fprintf(fp1, "%g\n", tag);
    for (int i=0; i<Xcut.size(); ++i) 
    {
      // TG below is old scalar version
      //double u = sol[ IJcut[i][0] ] + Scut[i] * (sol[ IJcut[i][1] ] - sol[ IJcut[i][0] ]);
      //fprintf(fp1, "%g\n", u);

      // TG now we can have any number of components for the solution	    
      for(int j = 0 ; j < solutionDimension; ++j)    
      {
	double u =  sol[ IJcut[i][0]*solutionDimension + j ] + Scut[i] * (sol[ IJcut[i][1]*solutionDimension + j ] - sol[ IJcut[i][0]*solutionDimension + j  ]);
	fprintf(fp1,"%g ", u);
	if(j+1<solutionDimension) fprintf(fp1," ");
      }
      fprintf(fp1,"\n");
    }
  }

  if (!gmsh && fp2) {
    fprintf(fp2, "%g\n", tag);
    for (int i=0; i<Ncut.size(); ++i)
    {
      //TG below is old scalar version	    
      //fprintf(fp2, "%g\n", sol[Ncut[i]]);
      for(int j = 0 ; j < solutionDimension; ++j)    
      {
	      fprintf(fp2, "%g ", sol[ Ncut[i]*solutionDimension+j ] );
	      if(j+1<solutionDimension) fprintf(fp2," ");
      }
      fprintf(fp2,"\n");
    }
  }

  if (gmsh) {
    fprintf(fp1, "View \"Slice2D %g\" {\n", tag);
    int i;
    for (i=0; i<Fcut.size(); ++i) {
      double s[3];
      for (int j=0; j<3; ++j) {
	int k = IJcut[ Fcut[i][j] ][0];
	int l = IJcut[ Fcut[i][j] ][1];
	s[j] = sol[k] + Scut[ Fcut[i][j] ] * (sol[l] - sol[k]);
      }
      fprintf(fp1, "ST(%g,%g,%g,%g,%g,%g,%g,%g,%g){%g,%g,%g};\n",
	      Xcut[ Fcut[i][0] ][0], Xcut[ Fcut[i][0] ][1], Xcut[ Fcut[i][0] ][2], 
	      Xcut[ Fcut[i][1] ][0], Xcut[ Fcut[i][1] ][1], Xcut[ Fcut[i][1] ][2],
	      Xcut[ Fcut[i][2] ][0], Xcut[ Fcut[i][2] ][1], Xcut[ Fcut[i][2] ][2],
	      s[0], s[1], s[2]);
      if (symmetry)
	fprintf(fp1, "ST(%g,%g,%g,%g,%g,%g,%g,%g,%g){%g,%g,%g};\n",
		xyzsym[0]*Xcut[ Fcut[i][0] ][0], xyzsym[1]*Xcut[ Fcut[i][0] ][1], xyzsym[2]*Xcut[ Fcut[i][0] ][2], 
		xyzsym[0]*Xcut[ Fcut[i][1] ][0], xyzsym[1]*Xcut[ Fcut[i][1] ][1], xyzsym[2]*Xcut[ Fcut[i][1] ][2],
		xyzsym[0]*Xcut[ Fcut[i][2] ][0], xyzsym[1]*Xcut[ Fcut[i][2] ][1], xyzsym[2]*Xcut[ Fcut[i][2] ][2],
		s[0], s[1], s[2]);
    }
    fprintf(fp1, "};\n");

    map<Face, pair<pair<int, int>,int> > gfaces;
    map<Face, pair<pair<int, int>,int> >::iterator it;
    for (i=0; i<Tcut.size(); ++i) {
      for (int k=0; k<4; ++k) {
	Face f;
	f[0] = tets[Tcut[i]][ Tet::face_def[k][0] ];
	f[1] = tets[Tcut[i]][ Tet::face_def[k][1] ];
	f[2] = tets[Tcut[i]][ Tet::face_def[k][2] ];
	f.reorder();
	it = gfaces.find(f);
	if (it == gfaces.end()) {
	  pair<int,int> pn(tets[Tcut[i]][ Tet::face2node_def[k] ], -(i + 1));
	  gfaces[f] = pair<pair<int,int>,int>(pn, k);
	}
	else
	  it->second.first.second = tets[Tcut[i]][ Tet::face2node_def[k] ];
      }
    }
    fprintf(fp2, "View \"Slice3D %g\" {\n", tag);
    for (it=gfaces.begin(); it!=gfaces.end(); ++it) {
      if (it->second.first.second < 0) {
	i = -(it->second.first.second + 1);
	int k = it->second.second;
	int f[3];
	f[0] = tets[Tcut[i]][ Tet::face_def[k][0] ];
	f[1] = tets[Tcut[i]][ Tet::face_def[k][2] ];
	f[2] = tets[Tcut[i]][ Tet::face_def[k][1] ];
	fprintf(fp2, "ST(%g,%g,%g,%g,%g,%g,%g,%g,%g){%g,%g,%g};\n",
		X[f[0]][0], X[f[0]][1], X[f[0]][2], X[f[1]][0], X[f[1]][1], X[f[1]][2], 
		X[f[2]][0], X[f[2]][1], X[f[2]][2], sol[f[0]], sol[f[1]], sol[f[2]]);
	if (symmetry)
	  fprintf(fp2, "ST(%g,%g,%g,%g,%g,%g,%g,%g,%g){%g,%g,%g};\n",
		  xyzsym[0]*X[f[0]][0], xyzsym[1]*X[f[0]][1], xyzsym[2]*X[f[0]][2], 
		  xyzsym[0]*X[f[1]][0], xyzsym[1]*X[f[1]][1], xyzsym[2]*X[f[1]][2], 
		  xyzsym[0]*X[f[2]][0], xyzsym[1]*X[f[2]][1], xyzsym[2]*X[f[2]][2], 
		  sol[f[0]], sol[f[1]], sol[f[2]]);
      }
    }
    fprintf(fp2, "};\n");
  }

}

//-------------------------------------------------------------------------------

void Domain::constructTraceOnSurface(char *prefix, char* suffix, SVec<double,3> &X, Vec<double> &sol, 
				     SVec<int,4> &Fcut, SVec<int,3> &IJcut, Vec<double> &Scut)
{

  Vec<double> data(Scut.size());

  int i;
  for (i=0; i<Scut.size(); ++i)
    data[i] = sol[ IJcut[i][0] ] + Scut[i] * (sol[ IJcut[i][1] ] - sol[ IJcut[i][0] ]);

  EdgeSet sedges;

  int numFaces = Fcut.size();

  SVec<int,3> faces2edges(numFaces);

  int faceNum = 0;
  for (i=0; i<numFaces; ++i) {
    int f[3] = {Fcut[i][0], Fcut[i][1], Fcut[i][2]};
    int type = Fcut[i][3];

    if ((type == 5) && (IJcut[ f[0] ][2] == 1) && 
	(IJcut[ f[1] ][2] == 1) && (IJcut[ f[2] ][2] == 1)) continue;

    faces2edges[faceNum][0] = sedges.find(f[0], f[1]);
    faces2edges[faceNum][1] = sedges.find(f[1], f[2]);
    faces2edges[faceNum][2] = sedges.find(f[2], f[0]);

    ++faceNum;
  }

  if (faceNum == 0) return;

  Vec<int> edgeTag(sedges.size());
  edgeTag = 0;

  faceNum = 0;

  for (i=0; i<numFaces; ++i) {
    int f[3] = {Fcut[i][0], Fcut[i][1], Fcut[i][2]};
    int type = Fcut[i][3];

    if ((type == 5) && (IJcut[ f[0] ][2] == 1) && 
	(IJcut[ f[1] ][2] == 1) && (IJcut[ f[2] ][2] == 1)) continue;
    
    edgeTag[ faces2edges[faceNum][0] ]++;
    edgeTag[ faces2edges[faceNum][1] ]++;
    edgeTag[ faces2edges[faceNum][2] ]++;

    ++faceNum;
  }

  SVec<int,4> bedges(sedges.size());

  int numBoundEdges = 0;

  EdgeSet::MapPair::iterator it_se;
  for (it_se = sedges.begin(); it_se != sedges.end(); ++it_se) {
    if (edgeTag[it_se->second] == 1) {
      bedges[numBoundEdges][0] = it_se->first.first;
      bedges[numBoundEdges][1] = it_se->first.second;
      ++numBoundEdges;
    }
  }

  int j, numCont = 0;

  for (i=0; i<numBoundEdges; ++i) {

    bedges[i][2] = numCont;

    int end = bedges[i][1];

    int next = -1;

    for (j=i+1; j<numBoundEdges; ++j) {
      if (bedges[j][0] == end) {
	next = j;
	break;
      }
      else if (bedges[j][1] == end) {
	bedges[j][1] = bedges[j][0];
	bedges[j][0] = end;
	next = j;
	break;
      }
    }

    if (next != -1) {
      for (int k=0; k<4; ++k) {
	int tmp = bedges[i+1][k];
	bedges[i+1][k] = bedges[next][k];
	bedges[next][k] = tmp;
      }
    }
    else
      ++numCont;

  }

  printf("       %d boundary edges with %d contour(s)\n", numBoundEdges, numCont);

  Vec<int> type(numCont);
  type = -10;

  SVec<int,2> index(numCont);
  index = 0;

  for (i=0; i<numBoundEdges; ++i)
    if (type[ bedges[i][2] ] == -10)
      type[ bedges[i][2] ] = IJcut[ bedges[i][0] ][2];

  for (i=0; i<numBoundEdges; ++i) {
    bedges[i][3] = type[ bedges[i][2] ];
    index[ bedges[i][2] ][1]++;
  }

  index[0][0] = 0;
  for (i=1; i<numCont; ++i)
    index[i][0] = index[i-1][0] + index[i-1][1];

  //#define BOEING

  int contNum = 0;

  for (j=0; j<numCont; ++j)  // for each contour
    { 

      if (type[j] == 1) 
	{
	  
	  char fname[MAXLINE];
	  sprintf(fname, "%s.%d.gnu%s", prefix, contNum, suffix);
	  FILE *fp = fopen(fname, "w");
	  
#ifdef BOEING
	  sprintf(fname, "%s.%d%s", prefix, contNum, suffix);
	  FILE *fp2 = fopen(fname, "w");
#endif
	  
	  int ix = 0;
	  int iy = 1;
	  double length = 0.0;
	  double surface = 0.0;
	  double force = 0.0;
	  
	  int nseg0 = index[j][0];
	  int nseg1 = index[j][0] + index[j][1];
	  
#ifdef BOEING
	  double xmin = X[ bedges[nseg0][0] ][0];
	  double xmax = X[ bedges[nseg0][0] ][0];
	  
	  for (i=nseg0+1; i<nseg1; ++i) 
	    {
	      xmin = min(xmin, X[ bedges[i][0] ][0]);
	      xmax = max(xmax, X[ bedges[i][0] ][0]);
	    }
	  xmin = min(xmin, X[ bedges[nseg1 - 1][0] ][0]);
	  xmax = max(xmax, X[ bedges[nseg1 - 1][0] ][0]);
#endif

	  for (i=nseg0; i<nseg1; ++i) 
	    {
	      int i0 = bedges[i][0];
	      int i1 = bedges[i][1];
	      Vec3D x0(X[i0]);
	      Vec3D x1(X[i1]);
	      double dist = sqrt((x1-x0) * (x1-x0));
	      length += dist;
	      surface += 0.5 * (x0[ix] + x1[ix]) * (x1[iy] - x0[iy]);
	      force += 0.5 * (data[i0] + data[i1]) * dist;
#ifdef BOEING
	      fprintf(fp2, "%e %e\n", (x0[0] - xmin)/(xmax - xmin), data[i0]);
#endif
	      //fprintf(fp, "%e %e %e %e\n", x0[0], x0[1], x0[2], data[i0]);  // TG before scalar only
	      fprintf(fp, "%e %e %e ", x0[0], x0[1], x0[2]);//TG don't forget the trailing space !! 
	      /////////////////

	      // TG now we can have any number of components for the solution	    
	      for(int j = 0 ; j < solutionDimension; ++j)    
		{
		  double u =  sol[ IJcut[i0][0]*solutionDimension + j ] + Scut[i0] * (sol[ IJcut[i0][1]*solutionDimension + j ] - sol[ IJcut[i0][0]*solutionDimension + j  ]);
		  fprintf(fp,"%e ", u);
		  if(j+1<solutionDimension) fprintf(fp," ");
		}
	      fprintf(fp,"\n");
	      
	      //////////////////

	    }
	  int i1 = bedges[nseg1 - 1][1];
	  Vec3D x1(X[i1]);
#ifdef BOEING
	  fprintf(fp2, "%e %e\n", (x1[0] - xmin)/(xmax - xmin), data[i1]);
#endif
	  //fprintf(fp, "%e %e %e %e\n", x1[0], x1[1], x1[2], data[i1]); // TG before
	  fprintf(fp, "%e %e %e ", x1[0], x1[1], x1[2]);//TG don't forget the trailing space !
	  // TG now we can have any number of components for the solution	    
	  for(int j = 0 ; j < solutionDimension; ++j)    
		{
		  double u =  sol[ IJcut[i1][0]*solutionDimension + j ] + Scut[i1] * (sol[ IJcut[i1][1]*solutionDimension + j ] - sol[ IJcut[i1][0]*solutionDimension + j  ]);
		  fprintf(fp,"%e ", u);
		  if(j+1<solutionDimension) fprintf(fp," ");
		}
	  fprintf(fp,"\n");
	      
	  //////////////////

	  fprintf(stderr, "       length=%e surface=%e force=%e\n", 
		  length, fabs(surface), 2.0*force);
	  
#ifdef BOEING
	  fclose(fp2);
#endif
	  fclose(fp);
	  
	  ++contNum;

    }

  }

}

//-------------------------------------------------------------------------------

ADT_TREE *Domain::setupADT(int &numPtsADT, ADT_PT **&ptsADT)
{

  double t0 = getCpuTime();

  fprintf(stdout, " ... building ADT for the points of <%s>", name); fflush(stdout);

  numPtsADT = nodes.size();

  ptsADT = (ADT_PT **) calloc(numPtsADT, sizeof(ADT_PT *));

  for (int i=0; i<numPtsADT; ++i) 
    ptsADT[i] = Alloc_ADT_PT(i, nodes[i], 3);

  double a1[3], b1[3];
  Compute_Root_BoundingBox(3, numPtsADT, ptsADT, a1, b1);

  ADT_TREE *rootADT = Create_ADT_TREE(3, numPtsADT, ptsADT, a1, b1);
 
  fprintf(stdout, " in %f s\n", getCpuTime()-t0);

  return rootADT;

}

//-------------------------------------------------------------------------------

void Domain::locatePoints(ADT_TREE *rootADT, int numPtsADT)
				
{

  double t0 = getCpuTime();

  fprintf(stdout, " ... locating the points in the elements of <%s>", name); 
  fflush(stdout);

  ADT_PT **ptsTmp = (ADT_PT **) calloc(numPtsADT, sizeof(ADT_PT *));
  ADT_TREE **stack = (ADT_TREE **) calloc(numPtsADT, sizeof(ADT_TREE *));

  for (int i=0; i<tets.size(); ++i) {
    tets[i].computeBoundingBox(nodes);
    tets[i].locatePoints(nodes, i, rootADT, ptsTmp, stack);
  }

  free(ptsTmp);
  free(stack);

  fprintf(stdout, " in %f s\n", getCpuTime()-t0);

}

//-------------------------------------------------------------------------------

int Domain::interpolateSolution(int numPtsADT, ADT_PT **ptsADT, 
				SVec<double,5> &sol, SVec<double,5> &interpSol)
{

  double t0 = getCpuTime();
  fprintf(stdout, " ... interpolating the solution in <%s>", name); fflush(stdout);

  int outPts = 0;
  for (int i=0; i<numPtsADT; ++i) {
    int numTet = ptsADT[i]->ptr;
    int numPt = ptsADT[i]->no;
    if (numTet != -1)
      tets[numTet].interpolateSolution(nodes, sol, ptsADT[i]->x, interpSol[numPt]);
    else
      ++outPts;
  }

  fprintf(stdout, " in %f s\n", getCpuTime()-t0);
  if (outPts)
    fprintf(stdout, " ... %d external points\n", outPts);

  return outPts;

}


//-------------------------------------------------------------------------------

void Domain::initializeOutsidePoints(int numPtsADT, ADT_PT** ptsADT, NodeSet& nodes2, 
				     SVec<double,5>& sol2, SVec<double,5>& sol)
{

#ifdef USE_ANN
  const int maxpts = 1;

  int nbnodes = nodes2.size();
  ANNpointArray bnodes = annAllocPts(nbnodes, 3);
  int i;
  for (i=0; i<nbnodes; ++i)
    bnodes[i] = nodes2[i];

  ANNkd_tree* kdtree = new ANNkd_tree(bnodes, nbnodes, 3);
  ANNidxArray index = new ANNidx[maxpts];
  ANNdistArray dist = new ANNdist[maxpts];

  for (i=0; i<numPtsADT; ++i) {
    if (ptsADT[i]->ptr == -1) {
      int numPt = ptsADT[i]->no;
      kdtree->annkSearch(nodes[numPt], maxpts, index, dist);
      for (int j=0; j<5; ++j) 
	sol[numPt][j] = sol2[ index[0] ][j];
    }
  }
#endif  

}

//-------------------------------------------------------------------------------

void Domain::checkOutsidePoints(int numPtsADT, ADT_PT **ptsADT, SVec<double,5> &sol)
{

  numberEdges();

  int *ja, *ia;
  createNodeToNodeConnectivity(ja, ia);

  fprintf(stdout, " ... fixing outside points in <%s>\n", name); fflush(stdout);

  for (int i=0; i<numPtsADT; ++i) {
    if (ptsADT[i]->ptr == -1) {
      int numPt = ptsADT[i]->no;
      int found = 0;
      for (int k=ia[numPt]; k<ia[numPt+1]; ++k) {
	ADT_PT *neigh = ptsADT[ ja[k] ];
	if (neigh->ptr != -1) {
	  for (int j=0; j<5; ++j) 
	    sol[numPt][j] = sol[neigh->no][j];
	  found = 1;
	  break;
	}
      }
      if (found == 0) {
	fprintf(stderr, " *** ERROR: no valid neighbor for point %d (%e, %e, %e)\n", 
		numPt, nodes[numPt][0], nodes[numPt][1], nodes[numPt][2]);
	char fname[500];
	sprintf(fname, "outpt%d", numPt+1);
	FILE* fp = fopen(fname, "w");
	fprintf(fp, "Nodes Node%d\n", numPt+1);
	fprintf(fp, "%d %e %e %e\n", 1, nodes[numPt][0], nodes[numPt][1], nodes[numPt][2]);
	int i = 1;
	int k;
	for (k=ia[numPt]; k<ia[numPt+1]; ++k) {
	  if (ja[k] != numPt)
	    fprintf(fp, "%d %e %e %e\n", ++i, nodes[ja[k]][0], nodes[ja[k]][1], nodes[ja[k]][2]);
	}
	fprintf(fp, "typedef 3 Line_2\n");
	fprintf(fp, "Elements Element%d using Node%d\n", numPt+1, numPt+1);
	for (k=1; k<i; ++k)
	  fprintf(fp, "%d 2 1 %d\n", k, k+1);
	fclose(fp);
      }
    }
  }

}

//-------------------------------------------------------------------------------

int Domain::createNodeToNodeConnectivity(int *&ja, int *&ia)
{

  int numNodes = nodes.size();

  int *numNeigh = (int *) alloca(sizeof(int) * numNodes);

  int i;
  for (i=0; i<numNodes; ++i) numNeigh[i] = 1;

  EdgeSet::MapPair::iterator it = edges.begin();
  EdgeSet::MapPair::iterator end = edges.end();

  while (it != end) {

    ++numNeigh[ it->first.first ];
    ++numNeigh[ it->first.second ];

    ++it;
  }

  int nnz = 0;
  for (i=0; i<numNodes; ++i) nnz += numNeigh[i];

  if (nnz != (2*edges.size() + numNodes)) {
    fprintf(stderr," *** ERROR: wrong number of nonzero blocks\n");
    exit(1);
  }

  // construction of ia

  ia = new int[numNodes+1];

  ia[0] = 0;

  for (i=0; i<numNodes; ++i) ia[i+1] = ia[i] + numNeigh[i];

  // construction of ja

  ja = new int[nnz];

  for (i=0; i<numNodes; ++i) {
    ja[ia[i]] = i;
    numNeigh[i] = 1;
  }

  it = edges.begin();

  while (it != end) {

    int i1 = it->first.first;
    int i2 = it->first.second;

    ja[ ia[i1] + numNeigh[i1] ] = i2;
    ++numNeigh[i1];

    ja[ ia[i2] + numNeigh[i2] ] = i1;
    ++numNeigh[i2];

    ++it;
  }
  
  return nnz;

}

//-------------------------------------------------------------------------------

void Domain::readGeometry(char *con, char *geo, int type)
{

  if (type == 0)
    readAsciiGeometry(geo);
  else if (type == 1)
    readBinaryGeometry(con, geo);
  else if (type == 2)
    readXpostGeometry(geo);

}

//-------------------------------------------------------------------------------

void Domain::readAsciiGeometry(char *geo)
{

  char fname[MAXLINE], line[MAXLINE];

  strcpy(fname, geo);

  FILE *fptr = fopen(fname, "r"); checkFileOpen(fptr, fname);

  int numNodes, numTets, numFaces;

  fgets(line, MAXLINE, fptr);
  sscanf(line, "%d %d %d", &numNodes, &numTets, &numFaces);

  fprintf(stdout, " ... reading topology from \'%s\'\n", fname);

  nodes.read(numNodes, fptr);
  tets.read(numTets, fptr);
  faces.read(numFaces, fptr);

  fclose(fptr);

}

//-------------------------------------------------------------------------------

void Domain::readBinaryGeometry(char *con, char *geo)
{

  const char* postopname = "POSTOPNAME";
  char* name = getenv(postopname);
  char* routine = "postOpObj";
  void *handle = dlopen(name, RTLD_NOW);

  char *msg = dlerror();
  if (msg) {
    fprintf(stderr," *** ERROR: dynamic loading of \'%s\': %s\n", name, msg);
    exit(-1);
  }

  PostOp **op = (PostOp **) dlsym(handle, routine);

  if (!op) {
    fprintf(stderr," *** ERROR: could not find \'%s\' in \'%s\'\n", routine, name);
    exit(-1);
  }
  else
    postop = *op;

  postop->initialize(con, geo);
  nodes.read(postop->getNumberOfNodes(), postop->getNodes());
  postop->releaseNodes();
  tets.read(postop->getNumberOfElements(), postop->getElements());
  postop->releaseElements();
  faces.read(postop->getNumberOfFaces(), postop->getFaces());
  postop->releaseFaces();

}

//-------------------------------------------------------------------------------

//------------------------------------------------------------------------------

struct SNode {
  double x[3];
};
struct SBFace {
  int nodes[3];
  int type; 
};
struct SElement {
  int nodes[4];
};

//------------------------------------------------------------------------------

void Domain::readXpostGeometry(char *geo)
{

  map<int,int> type_map;
  type_map[4] = 4; 
  type_map[5] = 5;

  stack<SNode> snodes;
  stack<SBFace> sbfaces;
  stack<SElement> selements;

  FILE* fp = fopen(geo, "r");
  if (!fp) {
    fprintf(stderr, "*** Error: could not open file \'%s\'\n", geo);
    exit(1);
  }

  int type_boco = 10;
  int type_read = 0;
  char line[MAXLINE], key1[MAXLINE], key2[MAXLINE];
  while (fgets(line, MAXLINE, fp) != 0) {
    sscanf(line, "%s", key1);
    bool skip = false;
    if (strcmp(key1, "Nodes") == 0) {
      skip = true;
      type_boco = 10;
      type_read = 1;
      sscanf(line, "%*s %s", key2);
    }
    else if (strcmp(key1, "Elements") == 0) {
      skip = true;
      type_boco = 10;
      type_read = 2;
      sscanf(line, "%*s %s", key2);
      /*
      for (int j=MIN_TYPE_BC; j<=MAX_TYPE_BC; ++j) {
        if (strncmp(key2, type_name[j], strlen(type_name[j])) == 0) {
          type_boco = j;
          break;
        }
      }
      */
    }
    else if (strcmp(key1, "typedef") == 0) {
      skip = true;
      int t;
      sscanf(line, "%*s %d %s", &t, key2);
      if (strcmp(key2, "Tri_3") == 0)
        type_map[t] = 4;
      else if (strcmp(key2, "Tetra_4") == 0)
        type_map[t] = 5;
    }

    if (!skip) {
      if (type_read == 1) {
        SNode sn;
        sscanf(line, "%*s %lf %lf %lf", &sn.x[0], &sn.x[1], &sn.x[2]);
        snodes.push(sn);
      }
      else if (type_read == 2) {
        int num, t;
        sscanf(line, "%d %d", &num, &t);
        map<int,int>::iterator it = type_map.find(t);
        if (it != type_map.end())
          t = it->second;

        if (t == 4) {
          SBFace sbf;
          sscanf(line, "%*s %*s %d %d %d", &sbf.nodes[0], &sbf.nodes[1], &sbf.nodes[2]);
          sbf.nodes[0] -= 1;
          sbf.nodes[1] -= 1;
          sbf.nodes[2] -= 1;
          sbf.type = type_boco;
          sbfaces.push(sbf);
        }
        else if (t == 5) {
          SElement se;
          sscanf(line, "%*s %*s %d %d %d %d", &se.nodes[0], &se.nodes[1], &se.nodes[2], &se.nodes[3]);
          se.nodes[0] -= 1;
          se.nodes[1] -= 1;
          se.nodes[2] -= 1;
          se.nodes[3] -= 1;
          selements.push(se);
        }
        else {
          fprintf(stderr, "*** Error: element %d has a wrong type (%d)\n", num, t);
          exit(1);
        }
      }
    }
    else {
      fprintf(stdout, "Found %s (%d)\n", key2, type_boco);
    }
  }

  fclose(fp);

  int nnodes = snodes.size();
  int nbfaces = sbfaces.size();
  int nelements = selements.size();

  fprintf(stdout, "\'%s\': %d nodes, %d boundary faces and %d tetrahedra\n",
          geo, nnodes, nbfaces, nelements);

  nodes.resize(nnodes);
  faces.resize(0);
  tets.resize(nelements);

  int i = 0;
  while (snodes.size()) {
    SNode sn = snodes.top();
    nodes[nnodes-i-1][0] = sn.x[0];
    nodes[nnodes-i-1][1] = sn.x[1];
    nodes[nnodes-i-1][2] = sn.x[2];
    snodes.pop();
    ++i;
  }

  i = 0;
  while (selements.size()) {
    SElement se = selements.top();
    tets[nelements-i-1][0] = se.nodes[0];
    tets[nelements-i-1][1] = se.nodes[1];
    tets[nelements-i-1][2] = se.nodes[2];
    tets[nelements-i-1][3] = se.nodes[3];
    selements.pop();
    ++i;
  }

}

//-------------------------------------------------------------------------------

bool Domain::readSolution(char* fname, FILE*& fp, PostOp::TypeSol type, int size,
			  int num, double& tag, double* sol)
{

  bool status = true;

  char line[MAXLINE];
  if (postop)
  {
	//	std::cout << "dim = " << postop->getNumberOfNodalSolutions(fname) << std::endl;
	//	std::cout << "postop" << num << " " << type << " " << tag << " " <<sol << "size:" << size<<endl;
    status = postop->getSolution(fname, num, type, tag, sol);
  }
  else {
    if (num == 0) {
      fp = fopen(fname, "r"); 
      if (!fp)
	return false;

      fgets(line, MAXLINE, fp); 
      fgets(line, MAXLINE, fp); 

      int _size;
      sscanf(line, "%d", &_size);
      if (_size != size) {
	fprintf(stderr, " *** ERROR: wrong size\n");
	exit(1);
      }
    }

    if (fgets(line, MAXLINE, fp) != 0) {
      sscanf(line, "%lf", &tag);
      for (int i=0; i<size; ++i) {
	fgets(line, MAXLINE, fp);
	sscanf(line, "%lf", &sol[i]);
      }
    }
    else
      status = false;

  }

  return status;

}

//-------------------------------------------------------------------------------

bool Domain::readPosition(char *prefix, int num, double& tag, SVec<double,3> &X)
{

  bool status = false;

  if (postop) {
    status = postop->getSolution(prefix, num, PostOp::ASCII, tag, 
				 reinterpret_cast<double*>(X.data()));
    X += nodes;
  }
  else
    fprintf(stderr, "*** Warning: read position not implemented\n");

  return status;

}

//-------------------------------------------------------------------------------

void Domain::writeSolution(char* fname, SVec<double,5>& sol)
{

  fprintf(stdout, " ... writing solution from <%s>\n", fname);

  BinFileHandler file(fname, "wb");
  int size = nodes.size();
  int dim = 5;
  int numRes = 1;
  double tag = 0.0;
  file.write(&size, 1);
  file.write(&dim, 1);
  file.write(&numRes, 1);
  file.write(&tag, 1);
  file.write(reinterpret_cast<double*>(sol.data()), size*dim);

}

//-------------------------------------------------------------------------------

void Domain::writeGmshGeometry(Vec<double> &sol)
{

  FILE *fpos = fopen("surface.pos", "w");
  fprintf(fpos, "View \"Surface - scalar\" {\n");

  double sbb[6] = {1.e99, 1.e99, 1.e99, -1.e99, -1.e99, -1.e99};

  for (int i=0; i<faces.size(); ++i)
    if (faces[i].getCode() < 0) {
      for (int j=0; j<3; ++j) {
	for (int k=0; k<3; ++k) {
	  sbb[k] = min(sbb[k], nodes[ faces[i][j] ][k]);
	  sbb[k+3] = max(sbb[k+3], nodes[ faces[i][j] ][k]);
	}
      }
      fprintf(fpos, "ST(%g,%g,%g,%g,%g,%g,%g,%g,%g){%g,%g,%g};\n",
	      nodes[faces[i][0]][0], nodes[faces[i][0]][1], nodes[faces[i][0]][2],
	      nodes[faces[i][1]][0], nodes[faces[i][1]][1], nodes[faces[i][1]][2],
	      nodes[faces[i][2]][0], nodes[faces[i][2]][1], nodes[faces[i][2]][2],
	      sol[faces[i][0]], sol[faces[i][1]], sol[faces[i][2]]);
    }
  fprintf(fpos, "};\n");

  FILE *fgeo = fopen("surface.geo", "w");
  fprintf(fgeo, "Point(1) = {%e,%e,%e,%e};\n", sbb[0], sbb[1], sbb[2], 1.0);
  fprintf(fgeo, "Point(2) = {%e,%e,%e,%e};\n", sbb[0], sbb[1], sbb[5], 1.0);
  fprintf(fgeo, "Point(3) = {%e,%e,%e,%e};\n", sbb[3], sbb[1], sbb[5], 1.0);
  fprintf(fgeo, "Point(4) = {%e,%e,%e,%e};\n", sbb[3], sbb[1], sbb[2], 1.0);
  fprintf(fgeo, "Point(5) = {%e,%e,%e,%e};\n", sbb[0], sbb[4], sbb[2], 1.0);
  fprintf(fgeo, "Point(6) = {%e,%e,%e,%e};\n", sbb[0], sbb[4], sbb[5], 1.0);
  fprintf(fgeo, "Point(7) = {%e,%e,%e,%e};\n", sbb[3], sbb[4], sbb[5], 1.0);
  fprintf(fgeo, "Point(8) = {%e,%e,%e,%e};\n", sbb[3], sbb[4], sbb[2], 1.0);
  fprintf(fgeo, "Line(1) = {1,4};\n"); 
  fprintf(fgeo, "Line(2) = {2,3};\n"); 
  fprintf(fgeo, "Line(3) = {6,7};\n"); 
  fprintf(fgeo, "Line(4) = {5,8};\n"); 
  fprintf(fgeo, "Line(5) = {1,5};\n");
  fprintf(fgeo, "Line(6) = {2,6};\n"); 
  fprintf(fgeo, "Line(7) = {3,7};\n");
  fprintf(fgeo, "Line(8) = {4,8};\n");
  fprintf(fgeo, "Line(9) = {1,2};\n");
  fprintf(fgeo, "Line(10) = {4,3};\n");
  fprintf(fgeo, "Line(11) = {8,7};\n");
  fprintf(fgeo, "Line(12) = {5,6};\n");

}

//------------------------------------------------------------------------------

void checkFileOpen(FILE *fptr, char *name)
{

  if (!fptr) {
    fprintf(stderr," *** ERROR: cannot open (%s)\n",name); 
    exit(1);
  }

}

//-------------------------------------------------------------------------------

void Domain::nodalValueExtract(char* nameSol, char* namePos, char* suffix, int globNodeNum)
{
  double cput0 = getCpuTime();
  double t0 = getTime();
  int nodenum = globNodeNum - 1;
  fprintf(stdout, "     Writing nodal values for node %d (%e, %e, %e)\n", globNodeNum, nodes[nodenum][0], nodes[nodenum][1], nodes[nodenum][2]);

  FILE *fsol1 = 0;
  FILE *fxpost = 0;;
 
  //TG
  int solSize;
  if(postop && nameSol[0] != 0)
  {
    solSize = nodes.size()*postop->getNumberOfNodalSolutions(nameSol);
    solutionDimension = postop->getNumberOfNodalSolutions(nameSol);
  }
  else
    solSize = nodes.size();
  
  Vec<double> sol(solSize);
  sol = 0.0;

  int i = 0;
  while (1) {
    double tag = 0.0;
    bool status = false;
    if (nameSol[0] != 0)
    {
      status = readSolution(nameSol, fxpost, PostOp::ASCII, sol.size(), i, tag, sol.data());
    }
 
    if (i != 0 && !status)
      break;

    if (i == 0 && nameSol[0] != 0) {
      char name[MAXLINE];
      sprintf(name, "nodalvalue.%d.%s.txt", globNodeNum, suffix);
      fsol1 = fopen(name, "w");
      if(solutionDimension == 1) 
        fprintf(fsol1, "#Scalar values for node %d with coordinates (%e %e %e) \n", globNodeNum, nodes[nodenum][0], nodes[nodenum][1], nodes[nodenum][2]);
      else
        if(solutionDimension == 3)
          fprintf(fsol1, "#Vector values for node %d with coordinates (%e %e %e) \n", globNodeNum, nodes[nodenum][0], nodes[nodenum][1], nodes[nodenum][2]);
      	else
        {
          fprintf(fsol1, "??? name under load for SliceNodes2D%s\n", suffix);
          fprintf(stdout,"Warning : This doesn't look like scalar or vector data, please check solution file header");
        }
    }
#ifdef ARL_DEBUG
    fprintf(stdout, "time slice %d (%e)\n", i, tag);
    fprintf(stdout, "\n\n value = %e\n", sol[nodenum*solutionDimension]);
#endif

    fprintf(fsol1, "%g ", tag);
    for(int j = 0 ; j < solutionDimension; ++j)
      fprintf(fsol1,"%g ", sol[nodenum*solutionDimension+j]);
    fprintf(fsol1,"\n");

    ++i;
  }

  if (fsol1) fclose(fsol1);
  if (fxpost) fclose(fxpost);

  fprintf(stdout, "Timings:\n");
  fprintf(stdout, "\t\tCPU Timing:\t%e s\n", getCpuTime()-cput0);
  fprintf(stdout, "\t\t    Timing:\t%e s\n", getTime()-t0);


}

//-------------------------------------------------------------------------------

// this function extracts the value of a field (pressure, 3D-velocity, etc) at a
// point in space given by its coordinates using a linear interpolation within the
// tet it belongs at each time-step.
void Domain::pointWiseValueExtract(char* nameSol, char* namePos, char* suffix, double sensor[3])
{
  double cput0 = getCpuTime();
  double t0 = getTime();
  double readingPosTime = 0.0;
  double readingSolTime = 0.0;
  fprintf(stdout, "     Writing point-wise values for sensor (%e, %e, %e)\n", sensor[0],sensor[1],sensor[2]);

  FILE *fsol1 = 0;
  FILE *fxpost = 0;;
 
  int solSize;
  if(postop && nameSol[0] != 0)
  {
    solSize = nodes.size()*postop->getNumberOfNodalSolutions(nameSol);
    solutionDimension = postop->getNumberOfNodalSolutions(nameSol);
  }
  else
    solSize = nodes.size();
  
  SVec<double,3> X(nodes.size());    X   = nodes;
  Vec<double> sol(solSize);          sol = 0.0;

  int tetNumber = -1;
  double localCoord[3];
  int i = 0;  // solution number (= time-iteration)
  while (1) { // iterate over all time-step solutions
  fprintf(stdout, " ... processing it %d\n", i); fflush(stdout);
    double tag = 0.0;
    bool status = false;
    if (namePos[0] != 0){
      // obtain new positions of nodes
      readingPosTime -= getTime();
      status = readPosition(namePos, i, tag, X); //reading position, not displacement (adding of nodes position at t=0 done in routine)
      readingPosTime += getTime();
    }
    if (nameSol[0] != 0){
      readingSolTime -= getTime();
      status = readSolution(nameSol, fxpost, PostOp::ASCII, sol.size(), i, tag, sol.data());
      readingSolTime += getTime();
    }
 
    if (i != 0 && !status)
      break;

    // write header for output file
    if (i == 0 && nameSol[0] != 0) {
      char name[MAXLINE];
      sprintf(name, "pointWiseValue.%s.txt", suffix);
      fsol1 = fopen(name, "w");
      fprintf(fsol1, "#Point-wise sensor is located at (%e, %e, %e)\n", sensor[0], sensor[1], sensor[2]);
      if(solutionDimension == 1) 
        fprintf(fsol1, "#Scalar values\n");
      else
        if(solutionDimension == 3)
          fprintf(fsol1, "#Vector values\n");
      	else
        {
          fprintf(fsol1, "??? name under load for SliceNodes2D%s\n", suffix);
          fprintf(stdout,"Warning : This doesn't look like scalar or vector data, please check solution file header");
        }
    }

    // find tet the sensor belongs to in the current configuration
    if(namePos[0] != 0 || i == 0){
      tetNumber = -1;
      for(int iTet=0; iTet<tets.size(); iTet++){
        tets[iTet].computeBoundingBox(X);
        if(tets[iTet].pointIn(X,sensor)){
          tetNumber = iTet;
          break;
        }
      }
    }
#ifdef ARL_DEBUG
    fprintf(stdout, "time slice %d (%e) ---> tetNumber = %d (%d %d %d %d)\n", i, tag, tetNumber, tets[tetNumber][0],tets[tetNumber][1],tets[tetNumber][2],tets[tetNumber][3]);
    fprintf(stdout, "(%e %e %e), (%e %e %e), (%e %e %e), (%e %e %e)\n",
        nodes[tets[tetNumber][0]][0],nodes[tets[tetNumber][0]][1],nodes[tets[tetNumber][0]][2],
        nodes[tets[tetNumber][1]][0],nodes[tets[tetNumber][1]][1],nodes[tets[tetNumber][1]][2],
        nodes[tets[tetNumber][2]][0],nodes[tets[tetNumber][2]][1],nodes[tets[tetNumber][2]][2],
        nodes[tets[tetNumber][3]][0],nodes[tets[tetNumber][3]][1],nodes[tets[tetNumber][3]][2]);
#endif

    // find local coordinates of the sensor in that tet and interpolate
    if(tetNumber >= 0){
      // local coordinates
      if(namePos[0] != 0 || i == 0) tets[tetNumber].barycentricCoord(X, sensor, localCoord);
      // linear interpolation
      double value[solutionDimension];
      tets[tetNumber].interpolate(sol, solutionDimension, localCoord, value);

      // output to file
      fprintf(fsol1, "%g ", tag);
      for(int j = 0 ; j < solutionDimension; ++j)
        fprintf(fsol1,"%g ", value[j]);
      fprintf(fsol1,"\n");
    }else
      fprintf(stdout, "*** Warning: sensor is out of domain at time slice %d (time=%e)\n", i, tag);

    ++i;
  }

  if (fsol1) fclose(fsol1);
  if (fxpost) fclose(fxpost);

  fprintf(stdout, "Timings:\n");
  fprintf(stdout, "\t\tCPU   Time           :\t%e s\n", getCpuTime()-cput0);
  fprintf(stdout, "\t\tTotal Time           :\t%e s\n", getTime()-t0);
  fprintf(stdout, "\t\tSolution Reading Time:\t%e s\n", readingSolTime);
  fprintf(stdout, "\t\tPosition Reading Time:\t%e s\n", readingPosTime);


}

//-------------------------------------------------------------------------------
//-------------------------------------------------------------------------------
